
(**
 * Hammer module: core utils, including
 * val hammer: pprog -> prog - Transforms parsed program into interpretable one
 * val display: prog -> unit - Pretty-prints a program
 * type 'a stream = Wait of 'a * 'a stream lazy_t
 * val variables: string list -> string stream - infinite stream of fresh 
   variable names
*)

open Ast
open Format
open Lazy

(** 
 * merge p1 p2 is the HOCore program were p1 and p2 were composed in parallel
*)
let merge ((m,l): prog) ((m',l'): prog) =
  (SMap.union (fun _ (s,r) (s',r') -> Some (s$s', r$r')) m m', var_union l l')


type 'a stream = Wait of 'a * 'a stream lazy_t

(** 
 * variables is a stream of different variable names, suitable for display
*)
let variables =
  let alphabet = ["t";"w";"v";"u";"z";"y";"x"] in
  let (a,l) = match alphabet with
    | []   -> ("",[])
    | a::l -> (a,l) in
  let rec f n () =
    let m = string_of_int n in
    List.fold_left (fun w x -> Wait (x^"_"^m, lazy w))
      (Wait (a^"_"^m, from_fun (f (n+1)))) l in
  if a = "" then f 0 () else
  List.fold_left (fun w x -> Wait (x, lazy w)) (Wait (a, from_fun (f 1))) l

let canonical =
  let rec f n =
    let m = string_of_int n in
    Wait ("_"^m, lazy (f (n+1))) in
  f 0

let rec instantiate (m, l) ((Wait (x, f)) as w) =
  let rec inst_send_list l' = function
    | [] -> l'
    | (Send s)::l -> inst_send_list ((DSend (instantiate s w))::l') l in
  let rec inst_rec_list l' = function
    | [] -> l'
    | (Rec g)::l ->
      inst_rec_list ((DRec (x, instantiate (g (channel_empty, var_singleton x))
                                            (force f) ))::l') l
  in (SMap.map (fun (s,r) -> (inst_send_list [] s, inst_rec_list [] r)) m,
      l)

exception NotReckognized

let rec num_of_prog ((m,_): disp_prog) =
  try let (_,r) = SMap.find "ACK_rep" m in
    match r with
    | [DRec (_,(m,_))] ->
      let (_,r) = SMap.find "Repeat" m in
      (match r with
       | [DRec (_,p)] -> 1 + num_of_prog p
       | _ -> raise NotReckognized)
    | _ -> raise NotReckognized
  with Not_found -> if SMap.mem "Zero" m then 0
      else raise NotReckognized

let rec trim_list m =
  try begin
    let (s_len,_) = SMap.find "Len" m in
    let (s_hd,_) = SMap.find "Hd" m in
    let (s_tl,_) = SMap.find "Tl" m in
    match (s_len, s_hd, s_tl) with
    | [DSend (len,_)], [DSend hd], [DSend (tl,_)]
      -> (match SMap.find "Repeat" len with
          | (_,[DRec (_,p)]) -> if  num_of_prog p = 0 then []
            else hd::(trim_list tl)
          | _ -> raise NotReckognized)
    | _ -> raise NotReckognized end
  with Not_found -> raise NotReckognized
      

let rec pretty_print_prog is_main fmt ((m, l): disp_prog) =
  if channel_size m = 0 && var_is_empty l then fprintf fmt "*@ " else
    let some_list = try Some (trim_list m) with NotReckognized -> None in
    let m = match some_list with
      | None -> m
      | Some _ -> SMap.remove "Len" (SMap.remove "Hd" (SMap.remove "Tl" m)) in
    let chan_size = channel_size m in
    let sep fmt () = fprintf fmt (if is_main then " |@.## " else " |@;<1 2>") in
    let void fmt () = fprintf fmt "@?" in
    let rec print_separated_list call_function separator next fmt = function
      | []   -> fprintf fmt ""
      | [x]  -> if next then fprintf fmt "%a%a" call_function x separator ()
        else fprintf fmt "%a" call_function x
      | x::l -> fprintf fmt "%a%a%a" call_function x separator ()
                  (print_separated_list call_function separator next) l in
    let rec pretty_print_send a fmt = function
      | DSend p -> fprintf fmt "%s@[<hov 2>< %a >@]" a (pretty_print_prog false) p in
    let rec pretty_print_rec a fmt = function
      | DRec (x,p) ->
        let action = fprintf fmt "%s(%s).@;<0 1>@[<hov 0>( %a )@]" a
            x (pretty_print_prog false) in
        if a = "Repeat"
        then (try fprintf fmt "%d" (num_of_prog p)
              with NotReckognized -> action p)
        else action p in
    ignore (var_fold (fun n x -> fprintf fmt "%s%a" x
                         (if n = 0 then void else sep)
                         (); (n-1))
              (var_size l - (if chan_size = 0 && some_list = None then 1
                             else 0)) l);
    begin match some_list with
      | None -> ()
      | Some l -> fprintf fmt "@[<hv 2>[ %a ]@]%a"
                    (print_separated_list (pretty_print_prog false)
                       (fun fmt _ -> fprintf fmt " ;@;<1 2>") false) l
                    (if chan_size <> 0 then sep else void) ()
    end;
    ignore (channel_fold (fun a (s,r) n ->
        print_separated_list (pretty_print_send a) sep (n<>0) fmt s;
        if not (s = [] || r = []) then fprintf fmt "%a" sep ();
        print_separated_list (pretty_print_rec  a) sep (n<>0) fmt r;
        n - 1 ) m (chan_size - 1))

let display p =
  Format.fprintf Format.std_formatter "%a@." (pretty_print_prog true)
    (instantiate p variables)


let rec hammer_aux env = function
  | PSend (a,p)  -> SMap.singleton a ([Send (hammer_aux env p)],[]), var_empty
  | PRec (a,x,p) -> (SMap.singleton a ([], [Rec (fun y ->
    hammer_aux (SMap.add x y env) p)]), var_empty)
  | PComp (p,q)  -> merge (hammer_aux env p) (hammer_aux env q)
  | PZero        -> SMap.empty, var_empty
  | PExe x       -> match (try Some (SMap.find x env) with Not_found -> None)
    with | Some p -> p
         | None   -> (SMap.empty, var_singleton x)

let hammer = hammer_aux (SMap.empty)


let rec make_count_map m = function
  | [] -> m
  | a::l -> let count = (try 1 + SMap.find a m with Not_found -> 1) in
    make_count_map (SMap.add a count m) l

let rec compare_prog p1 p2 =

  let rec compare_aux (Wait (x,f) as w) ((m1,l1): prog) ((m2,l2): prog) =
    let y = if l1 == l2 then 0 else
        let l1' = make_count_map SMap.empty l1 in
        let l2' = make_count_map SMap.empty l2 in
        SMap.compare (fun x y -> x - y) l1' l2' in
    if y  <>  0 then y else
    if m1 == m2 then 0 else

      let compare_send_list s1 s2 =
        if s1 == s2 then 0 else
          let y = List.length s1 - List.length s2 in
          if y <> 0 then y else
            let sort_send_list =
              List.sort (fun (Send p) (Send p') -> compare_aux w p p') in
            let s1' = sort_send_list s1 in
            let s2' = sort_send_list s2 in
            List.fold_left2 (fun y (Send p1) (Send p2) ->
                if y <> 0 then y else compare_aux w p1 p2) 0 s1' s2' in

      let compare_receive_list r1 r2 =
        if r1 == r2 then 0 else
          let y = List.length r1 - List.length r2 in
          if y <> 0 then y else
            let sort_receive_list r =
              let r = List.map (fun (Rec g) ->
                  g (channel_empty, var_singleton x)) r in
              List.sort (fun p p' -> compare_aux (force f) p p') r in
            let r1' = sort_receive_list r1 in
            let r2' = sort_receive_list r2 in
            List.fold_left2 (fun y p1 p2 ->
                if y <> 0 then y else compare_aux (force f) p1 p2) 0 r1' r2' in

      SMap.compare (fun (s1,r1) (s2,r2) ->
          let y = compare_send_list s1 s2 in
          if y <> 0 then y else compare_receive_list r1 r2) m1 m2 in

  compare_aux canonical p1 p2

and compare_send ((Send p1) as s1) ((Send p2) as s2) =
  if s1 == s2 then 0 else compare_prog p1 p2

and compare_receive ((Rec f1) as r1) ((Rec f2) as r2) =
  if r1 == r2 then 0 else compare_prog (f1 (channel_empty, var_singleton "_"))
      (f2 (channel_empty, var_singleton "_"))

module ProgMap = Map.Make(struct
    type t = prog
    let compare = compare_prog end)

module PSet = Set.Make(struct
    type t = prog
    let compare = compare_prog end)
