open Ast
open Format
open Hammer
open Scanf

let wait _ = Scanf.bscanf Scanf.Scanning.stdin "%s\n" ignore

let extract_unique l =
  let rec extract_unique_aux l' before = function
    | [] -> l'
    | a::l -> extract_unique_aux ((a, before $ l) :: l') (a :: before) l in
  extract_unique_aux [] [] l


let one_step ((m,_) as p: prog) =
  let one_channel a (s,r) tail =
    let s = extract_unique s in
    let r = extract_unique r in
    let operate tail (Send send, slist) =
      let rec op_aux tail (Rec g, rlist) =
        PSet.add (merge (g send)
           (update_channel a p (slist, rlist)) ) tail
      in
      in_chan_fold op_aux tail r 
    in
    in_chan_fold operate tail s
  in
  channel_fold one_channel m PSet.empty

let display_one t =
  display t; printf "@."


let visualize max p =
  let rec falcon max p =
    if PSet.is_empty p then printf "All terminal states found@." else
    if max = 0 then
      (printf "Some reductions left, for instance: %a@."
         (pretty_print_prog true) (instantiate (PSet.choose p) variables)) else
      let p' = PSet.fold (fun x l -> let q = one_step x in
                           if PSet.is_empty q then (display_one x; l)
                           else PSet.union q l) p PSet.empty in
      falcon (max-1) p' in
  falcon max (PSet.singleton p)

type 'a choice = Zero | One of 'a | Choice of 'a list

let choice_random ((m,_) as p : prog) =
  let one_channel a (s,r) tail =
    let s = extract_unique s in
    let r = extract_unique r in
    let operate tail (Send send, slist) =
      let rec op_aux tail (Rec g, rlist) =
        PSet.add (merge (g send)
           (update_channel a p (slist, rlist)) ) tail
      in
      in_chan_fold op_aux tail r 
    in
    in_chan_fold operate tail s
  in
  let x = channel_fold one_channel m PSet.empty in
  match PSet.elements x with
  | []  -> (Zero: prog choice)
  | [a] -> One a
  | l   -> Choice l


let rec machine_small_steps p =
  printf "!@.";
  match choice_random p with
  | Zero -> printf "%a@.Final state reached@."
              (pretty_print_prog true) (instantiate p variables)
  | One x -> printf "Current state:@."; display_one p;
    wait (); machine_small_steps x
  | Choice l ->
    printf "Current state:@."; display_one p;
    printf "Multiple possible reductions, choosing one.@.";
    let x = Random.int (List.length l) in wait (); machine_small_steps (List.nth l x)

let rec machine_steps p =
  match choice_random p with
  | Zero -> printf "%a@.Final state reached@."
              (pretty_print_prog true) (instantiate p variables)
  | One x -> machine_steps x
  | Choice l -> printf "Current state:@."; display_one p;
    printf "Multiple possible reductions, choosing one.@.";
    let x = Random.int (List.length l) in wait (); machine_steps (List.nth l x)

let rec machine_random max p =
  if max = 0 then (display p; printf "Not a terminal state@.") else begin
    match choice_random p with
    | Zero -> printf "%a@.Final state reached@."
                (pretty_print_prog true) (instantiate p variables)
    | One x -> machine_random (max-1) x
    | Choice l -> 
    let x = Random.int (List.length l) in machine_random (max-1) (List.nth l x)
  end


type select_choice = Nope | Unique of (send * receive * prog)
                   | Choice of (send * receive * prog)

let ask message f =
  printf "%s @?" message;
  bscanf Scanning.stdin "%u\n" f

let oracle_ask ((m,l) : prog) =
  let rec regroup a old l' = function
    | [] -> l'
    | x::l -> regroup a (x::old) ((a,x,old $ l)::l') l in
  let trim a ll (ret_s,ret_r) =
    match ll with
    | [], _ | _, [] -> (ret_s,ret_r)
    | [s], [r] -> ((a,s,[])::ret_s, (a,r,[])::ret_r)
    | ls, lr -> (regroup a [] ret_s ls, regroup a [] ret_r lr) in
  match SMap.fold trim m ([],[]) with
  | [],[] -> Nope
  | [(a,s,ls)],[(_,r,lr)] ->
    let m = if ls = [] && lr = [] then SMap.remove a m
      else SMap.add a (ls,lr) m in                                 
    Unique (s,r, (m,l))
  | ls, lr ->
    printf "Current state:@."; display_one (m,l);
    let propose message to_print l =
      List.iteri (fun i x -> printf "- n°%d: %a@." i to_print x) l;
      ask message (fun n -> n)in
    let one_print_send fmt (a,s,_) =
      pretty_print_prog true fmt (instantiate (SMap.singleton a ([s],[]),[])
                                    variables) in
    let one_print_recv fmt (a,r,_) =
      pretty_print_prog true fmt (instantiate (SMap.singleton a ([],[r]),[])
                                    variables) in
    let rec check_valid () =
      let i_s = propose  "Which emission? " one_print_send ls in
      let i_r = propose "Which reception? " one_print_recv lr in
      let (a,s,ls) = List.nth ls i_s in
      let (b,r,lr) = List.nth lr i_r in
      if a<>b then
        (printf "Emission and reception must be on the same channel.@.@.";
         check_valid ())
      else (printf "@.----------@.@."; (a,s,ls,r,lr)) in
    let (a,s,ls,r,lr) = check_valid () in
    let m = if ls = [] && lr = [] then SMap.remove a m
      else SMap.add a (ls,lr) m in
    Choice(s,r,(m,l))

let rec machine_ask p =
  match oracle_ask p with
  | Nope -> printf "%a@.Final state reached@."
              (pretty_print_prog true) (instantiate p variables)
  | Unique (Send s, Rec r,p) | Choice (Send s, Rec r,p)
    -> machine_ask (merge (r s) p)

let rec machine_ask_all p =
  match oracle_ask p with
  | Nope -> printf "%a@.Final state reached@."
              (pretty_print_prog true) (instantiate p variables)
  | Unique (Send s, Rec r,p)
    -> printf "Current state:@."; display_one p;
    machine_ask (merge (r s) p)
  | Choice (Send s, Rec r,p) -> machine_ask (merge (r s) p)
